// XMLStack.h
// Author: Scott Moore
// Permission to freely use, copy, modify, and distribute this source code 
// is hereby granted.  No representations are made about the
// suitability of this software for any purpose.  It is provided
// "as is" without express or implied warranty.
#ifndef _XMLSTACK_H
#define _XMLSTACK_H

#include <string>
#include <vector>
#include <map>
#include <exception>
#include <algorithm>
#include <sstream>
#include <iomanip>
#include "ConvertEncoding.h"

// The comments in this file are formatted to work with doxygen.
// See http://www.doxygen.org for details.

#define _CE ConvertEncoding

// Forward declarations
template<class _E, class _Tr, class _A> class XMLStackDocument;
template<class _E, class _Tr, class _A> class XMLStackAttribute;



/// Exception thrown by the XMLStack* classes.
class XMLStackException : public std::exception
{
public:
  /// Constructor
  XMLStackException(const std::string & Text) : m_Msg(Text) {}

  /// Destructor
  virtual ~XMLStackException() {}

  /// Exception error message
  virtual const char *what() const { return m_Msg.c_str(); }
private:
  std::string m_Msg;
};




/// Abstract base class for all node types
template<class _E, class _Tr = std::char_traits<_E>,  class _A = std::allocator<_E> >
class XMLStackNode
{
public:
  typedef XMLStackNode<_E, _Tr, _A> _XMLStackNode;
  typedef XMLStackDocument<_E, _Tr, _A> _XMLStackDocument;
  typedef XMLStackAttribute<_E, _Tr, _A> _XMLStackAttribute;
  typedef std::basic_string<_E, _Tr, _A> _XMLStackString;

  /** Constructor
   *
   *  \param Doc The XMLStackDocument this class instance belongs to
   *  \param Name The qualified name of the node.  Qualified node names
   *         include the namespace prefix (if any), a colon ":" and then
   *         the node name (local part).  If the node does not belong
   *         to a specific namespace, then only the local name needs to
   *         be supplied.
   */
  XMLStackNode(_XMLStackDocument *     Doc,
               const _XMLStackString & Name) : m_Doc(Doc), m_Name(Name) {}
  
  /// Destructor
  virtual ~XMLStackNode() {}

  /// Returns the qualified node name (namespace prefix and local name)
  const _XMLStackString & getNodeName(void) const { return m_Name; }
  
  /// Returns the namespace prefix, if there is one
  _XMLStackString getPrefix(void) const
  {
    _XMLStackString::size_type Pos;
    _XMLStackString            Prefix;

    Pos = m_Name.find(':');
    if (Pos != _XMLStackString::npos)
    {
      Prefix.assign(m_Name.begin(), m_Name.begin() + Pos);
    }
    return Prefix;
  }

  /// Returns the node's local name without the prefix
  _XMLStackString getLocalName(void) const
  {
    _XMLStackString::size_type Pos;
    _XMLStackString            Name;

    Pos = m_Name.find(':');
    if (Pos != _XMLStackString::npos)
    {
      Name.assign(m_Name, Pos + 1, _XMLStackString::npos);
    }
    else
    {
      return m_Name;
    }
    return Name;
  }

  /// Creates an attribute node
  virtual _XMLStackAttribute & createAttribute
    (const _XMLStackString & Name,
     const _XMLStackString & NamespaceURI = _XMLStackString())
  {
    throw XMLStackException("ERROR: Node type does not support attributes.");
  }

  /// Returns an existing attribute node based on the qualified name
  virtual _XMLStackAttribute & getAttribute(const _XMLStackString & Name)
  {
    throw XMLStackException("ERROR: Node type does not support attributes.");
  }

  /// Returns true if a node with the given qualified name exists
  virtual bool hasAttribute(const _XMLStackString & Name) const
  {
    throw XMLStackException("ERROR: Node type does not support attributes.");
  }

  /** Convenience function to assign a "value" to a node.
   *
   *  These functions are intended to be defined by derived classes,
   *  where appropriate, and to assign the values passed in
   *  as befits that node type.  For instance, element nodes
   *  would push and pop a child text node with the given value.
   *
   *  When these calls don't make sense for a derived class,
   *  they shouldn't implement them.  The default behavior is to throw
   *  an XMLStackException().
   */
  virtual void setValue(const _XMLStackString & Value)
  {
    throw XMLStackException("ERROR: setValue() not implemented for this"
      " node type");
  }
  virtual void setValue(unsigned long Value)
  {
    throw XMLStackException("ERROR: setValue() not implemented for this"
      " node type");
  }
  virtual void setValue(long Value)
  {
    throw XMLStackException("ERROR: setValue() not implemented for this"
      " node type");
  }
  virtual void setValue(unsigned int Value)
  {
    throw XMLStackException("ERROR: setValue() not implemented for this"
      " node type");
  }
  virtual void setValue(int Value)
  {
    throw XMLStackException("ERROR: setValue() not implemented for this"
      " node type");
  }
  virtual void setValue(unsigned short Value)
  {
    throw XMLStackException("ERROR: setValue() not implemented for this"
      " node type");
  }
  virtual void setValue(short Value)
  {
    throw XMLStackException("ERROR: setValue() not implemented for this"
      " node type");
  }
  virtual void setValue(signed char Value)
  {
    throw XMLStackException("ERROR: setValue() not implemented for this"
      " node type");
  }
  virtual void setValue(char Value)
  {
    throw XMLStackException("ERROR: setValue() not implemented for this"
      " node type");
  }
  virtual void setValue(unsigned char Value)
  {
    throw XMLStackException("ERROR: setValue() not implemented for this"
      " node type");
  }
  virtual void setValue(double Value, int Scale = 6)
  {
    throw XMLStackException("ERROR: setValue() not implemented for this"
      " node type");
  }
  virtual void setValue(float Value, int Scale = 6)
  {
    throw XMLStackException("ERROR: setValue() not implemented for this"
      " node type");
  }

  /// Convenience operator for setValue()
  _XMLStackNode & operator=(const _XMLStackString & Value)
  {
    setValue(Value);
    return *this;
  }
  /// Convenience operator for setValue()
  _XMLStackNode & operator=(unsigned long & Value)
  {
    setValue(Value);
    return *this;
  }
  /// Convenience operator for setValue()
  _XMLStackNode & operator=(long Value)
  {
    setValue(Value);
    return *this;
  }
  /// Convenience operator for setValue()
  _XMLStackNode & operator=(unsigned int Value)
  {
    setValue(Value);
    return *this;
  }
  /// Convenience operator for setValue()
  _XMLStackNode & operator=(int Value)
  {
    setValue(Value);
    return *this;
  }
  /// Convenience operator for setValue()
  _XMLStackNode & operator=(unsigned short Value)
  {
    setValue(Value);
    return *this;
  }
  /// Convenience operator for setValue()
  _XMLStackNode & operator=(short Value)
  {
    setValue(Value);
    return *this;
  }
  /// Convenience operator for setValue()
  _XMLStackNode & operator=(signed char Value)
  {
    setValue(Value);
    return *this;
  }
  /// Convenience operator for setValue()
  _XMLStackNode & operator=(char Value)
  {
    setValue(Value);
    return *this;
  }
  /// Convenience operator for setValue()
  _XMLStackNode & operator=(unsigned char Value)
  {
    setValue(Value);
    return *this;
  }
  /// Convenience operator for setValue()
  _XMLStackNode & operator=(double Value)
  {
    setValue(Value);
    return *this;
  }
  /// Convenience operator for setValue()
  _XMLStackNode & operator=(float Value)
  {
    setValue(Value);
    return *this;
  }
protected:
  virtual bool childNodePushed(void) { return false; }
  virtual void popped(void) { }
  virtual _XMLStackNode * clone(void) const = 0;

  typedef std::basic_stringstream<_E, _Tr, _A> _XMLStackStringStream;

  _XMLStackString convertValue(unsigned long Value)
  {
    _XMLStackStringStream Stream;
    Stream << Value;
    return Stream.str();
  }

  _XMLStackString convertValue(long Value)
  {
    _XMLStackStringStream Stream;
    Stream << Value;
    return Stream.str();
  }

  _XMLStackString convertValue(unsigned int Value)
  {
    _XMLStackStringStream Stream;
    Stream << Value;
    return Stream.str();
  }

  _XMLStackString convertValue(int Value)
  {
    _XMLStackStringStream Stream;
    Stream << Value;
    return Stream.str();
  }

  _XMLStackString convertValue(unsigned short Value)
  {
    _XMLStackStringStream Stream;
    Stream << Value;
    return Stream.str();
  }

  _XMLStackString convertValue(short Value)
  {
    _XMLStackStringStream Stream;
    Stream << Value;
    return Stream.str();
  }

  _XMLStackString convertValue(signed char Value)
  {
    _XMLStackStringStream Stream;
    Stream << Value;
    return Stream.str();
  }

  _XMLStackString convertValue(char Value)
  {
    _XMLStackStringStream Stream;
    Stream << Value;
    return Stream.str();
  }

  _XMLStackString convertValue(unsigned char Value)
  {
    _XMLStackStringStream Stream;
    Stream << Value;
    return Stream.str();
  }

  _XMLStackString convertValue(double Value, int Scale)
  {
    _XMLStackStringStream Stream;
    Stream << std::fixed << std::setprecision(Scale) << Value;
    return Stream.str();
  }

  _XMLStackString convertValue(float Value, int Scale)
  {
    _XMLStackStringStream Stream;
    Stream << std::fixed << std::setprecision(Scale) << Value;
    return Stream.str();
  }

  // Appends the passed in XML to the document's XML buffer.
  // NOTE: It is important to use this method instead of
  // GetBuffer().append() because this method provides
  // better performance by minimizing memory allocations
  // and copying.
  void appendXML(const _XMLStackString & XML)
  {
    m_Doc->appendXML(XML);
  }

  // Returns a reference to the XML buffer held by the XMLStackDocument
  inline _XMLStackString & getBuffer(void) { return m_Doc->m_XML; }

  _XMLStackDocument * m_Doc;
  _XMLStackString     m_Name;

private:

  friend class XMLStackDocument<_E, _Tr, _A>;
};


/** This node derived class creates element nodes in the XML.
 *
 *  Element nodes are the most common node type.  They are also
 *  the only node type that supports attribute nodes.
 */
template<class _E, class _Tr = std::char_traits<_E>,  class _A = std::allocator<_E> >
class XMLStackElement : public XMLStackNode<_E, _Tr, _A>
{
public:
  typedef XMLStackElement<_E, _Tr, _A> _XMLStackElement;

  /** Constructor which takes a node name
   *
   *  The element name must conform to the rules for valid element
   *  names.  Only some basic checking is done to make sure the
   *  name is valid, and only when the _DEBUG preprocessor symbol is
   *  defined.
   *
   *  \param Doc The XMLStackDocument this element belongs to
   *  \param Name The name of the element
   *
   *  \exception XMLStackException Thrown if Name is invalid
   */
  XMLStackElement(_XMLStackDocument *     Doc,
                  const _XMLStackString & Name)
   : _XMLStackNode(Doc, Name)
  {
    m_StartTagBegin = 0;
    m_StartTagEnd   = 0;
    m_Dirty         = false;
    m_NSDefined     = false;

#ifdef _DEBUG
    checkName();
#endif
  }

  /** Constructor which takes both a node name and a NamespaceURI
   *
   *  The element name must conform to the rules for valid element
   *  names.  Only some basic checking is done to make sure the
   *  name is valid, and only when the _DEBUG preprocessor symbol is
   *  defined.
   *
   *  If a NamespaceURI is provided, no checking is performed to see
   *  if it was declared in an ancestor element; it will be written
   *  out regardless.  If you want the element in a namespace but don't
   *  want the xmlns declaration again, only prefix the element name
   *  with the namespace qualifier (e.g., "NamespaceQualifier:ElementName").
   *  Don't provide the NamespaceURI.
   *
   *  \param Doc The XMLStackDocument this element belongs to
   *  \param Name The name of the element, along with any namespace prefix
   *         unless the node should declare a default namespace.
   *  \param NamespaceURI String defining the namespace URI.  Even an empty
   *         string will declare an empty namespace.  If the Name
   *         parameter does not have a prefix, this is treated as the
   *         default namespace. 
   *
   *  \exception XMLStackException Thrown if Name is invalid
   */
  XMLStackElement(_XMLStackDocument *     Doc,
                  const _XMLStackString & Name,
                  const _XMLStackString & NamespaceURI)
   : _XMLStackNode(Doc, Name), m_NamespaceURI(NamespaceURI)
  {
    m_StartTagBegin = 0;
    m_StartTagEnd   = 0;
    m_Dirty         = false;
    m_NSDefined     = true;

#ifdef _DEBUG
    checkName();
#endif
  }

  /// Destructor
  virtual ~XMLStackElement()
  {
    // Free up any attributes
    for (AttributesVec::iterator Iter = m_AttributesVec.begin();
         Iter != m_AttributesVec.end(); ++Iter)
    {
      delete (*Iter);
    }
  }

  /// Returns the URI for the namespace (if any)
  const _XMLStackString & getNamespaceURI(void) const
  {
    return m_NamespaceURI;
  }

  /** Convenience method that creates a text node with Value
   *
   *  This method pushes a text node onto the stack, assigns Value to it,
   *  and then pops it back off.
   *
   *  \note One of the side effects of this operation is that multiple
   *        assignments actually operate like additions.  Every assignment
   *        causes another text node to be pushed and popped, thus
   *        the text will get appended instead of replacing the previous
   *        assignment value.
   *
   *  \param Value The string that will be assigned to the text node
   */
  virtual void setValue(const _XMLStackString & Value)
  {
    m_Doc->pushText(Value);
    m_Doc->pop();
  }
  virtual void setValue(unsigned long Value) { setValue(convertValue(Value)); }
  virtual void setValue(long Value) { setValue(convertValue(Value)); }
  virtual void setValue(unsigned int Value) { setValue(convertValue(Value)); }
  virtual void setValue(int Value) { setValue(convertValue(Value)); }
  virtual void setValue(unsigned short Value) { setValue(convertValue(Value)); }
  virtual void setValue(short Value) { setValue(convertValue(Value)); }
  virtual void setValue(signed char Value) { setValue(convertValue(Value)); }
  virtual void setValue(char Value) { setValue(convertValue(Value)); }
  virtual void setValue(unsigned char Value) { setValue(convertValue(Value)); }
  virtual void setValue(double Value, int Scale = 6)
  { 
    setValue(convertValue(Value, Scale)); 
  }
  virtual void setValue(float Value, int Scale = 6)
  { 
    setValue(convertValue(Value, Scale)); 
  }

  /** Creates an attribute node for this element node
   *
   *  \note For better performance, you should create all your attributes
   *        and assign them values BEFORE pushing any other nodes onto
   *        the stack.  If you create or modify an attribute after other
   *        nodes have been pushed and popped on this element, then the
   *        previously written XML will have to be replaced, thus requiring
   *        a potentially expensive string copy.
   *
   *  \param Name The name of the attribute, with any namespace prefix
   *         included if present.
   *  \param NamespaceURI String defining the namespace URI. If specified,
   *         the node is created in the context of the namespaceURI
   *         with the prefix specified on the node name.  If the Name
   *         parameter does not have a prefix, this is treated as the
   *         default namespace. 
   *
   *  \exception XMLStackException Thrown if the attribute name is invalid or
   *             if an attribute already exists with that name
   */
  virtual _XMLStackAttribute & createAttribute
    (const _XMLStackString & Name,
     const _XMLStackString & NamespaceURI = _XMLStackString())
  {
    if (m_Attributes.find(Name) == m_Attributes.end())
    {
      // We use the vector to maintain the attributes in the order they
      // are created
      m_AttributesVec.push_back
        (new _XMLStackAttribute(m_Doc, Name, NamespaceURI));

      // We use the map for fast lookup
      m_Attributes[Name] = m_AttributesVec.back();
      m_Dirty = true;
      return *m_AttributesVec.back();
    }
    std::string Msg;

    Msg = "ERROR: Element node ";
    Msg += _CE(m_Name.c_str());
    Msg += " already has an Attribute node named ";
    Msg += _CE(Name.c_str());
    Msg += ".";
    throw XMLStackException(Msg);
  }

  /** Returns the attribute node with the specified name.
   *
   *  \param Name The qualified name of the attribute node
   *
   *  \exception XMLStackException Thrown if the attribute node doesn't exist
   */
  virtual _XMLStackAttribute & getAttribute(const _XMLStackString & Name)
  {
    Attributes::iterator Iter = m_Attributes.find(Name);

    if (Iter != m_Attributes.end())
    {
      return *(*Iter).second;
    }
    std::string Msg;

    Msg = "ERROR: Element node ";
    Msg += _CE(m_Name.c_str());
    Msg += " does not have an Attribute node named ";
    Msg += _CE(Name.c_str());
    Msg += ".";
    throw XMLStackException(Msg);
  }

  /** Returns true if the attribute node exists, false otherwise
   *
   *  \param Name The qualified name of the attribute node
   */
  virtual bool hasAttribute(const _XMLStackString & Name) const
  {
    if (m_Attributes.find(Name) != m_Attributes.end())
    {
      return true;
    }
    return false;
  }
  
  /// Convenience operator for setValue()
  _XMLStackElement & operator=(const _XMLStackString & Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackElement & operator=(unsigned long & Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackElement & operator=(long Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackElement & operator=(unsigned int Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackElement & operator=(int Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackElement & operator=(unsigned short Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackElement & operator=(short Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackElement & operator=(signed char Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackElement & operator=(char Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackElement & operator=(unsigned char Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackElement & operator=(double Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackElement & operator=(float Value)
  {
    setValue(Value);
    return *this;
  }


protected:
  virtual bool childNodePushed(void)
  {
    writeStartTag();
    return true;
  }

  typedef std::map<_XMLStackString, _XMLStackAttribute *> Attributes;
  typedef std::vector<_XMLStackAttribute *> AttributesVec;

  _XMLStackString m_NamespaceURI;
  Attributes      m_Attributes;
  AttributesVec   m_AttributesVec;
  size_t          m_StartTagBegin;
  size_t          m_StartTagEnd;
  bool            m_Dirty;
  bool            m_NSDefined;

  void writeStartTag(bool EndTag = false)
  {
    if ((m_StartTagBegin + m_StartTagEnd) == 0 || m_Dirty)
    {
      _XMLStackString Tag;

      Tag = m_Doc->VALS.LT;
      Tag += m_Name;

      // A namespace was defined, so we will need to output the definition
      if (m_NSDefined)
      {
        _XMLStackString Prefix = getPrefix();

        if (Prefix.length() > 0)
        {
          Tag += m_Doc->VALS.S_XMLNS_C;
          Tag += Prefix;
          Tag += m_Doc->VALS.EQUAL + m_Doc->VALS.QUOTE;
          Tag += m_NamespaceURI + m_Doc->VALS.QUOTE;
        }
        else
        {
          Tag += m_Doc->VALS.SPACE + m_Doc->VALS.XMLNS + m_Doc->VALS.EQUAL;
          Tag += m_Doc->VALS.QUOTE + m_NamespaceURI + m_Doc->VALS.QUOTE;
        }
      }

      // Write out all the attributes
      for (AttributesVec::iterator Iter = m_AttributesVec.begin();
           Iter != m_AttributesVec.end(); ++Iter)
      {
        Tag += m_Doc->VALS.SPACE;
        Tag += (*Iter)->xml();
        (*Iter)->setDirty(false);
      }

      if (EndTag)
      {
        Tag += m_Doc->VALS.SLASH + m_Doc->VALS.GT;
      }
      else
      {
        Tag += m_Doc->VALS.GT;
      }

      if ((m_StartTagBegin + m_StartTagEnd) == 0)
      {
        // Never written out before, this is the most efficient
        m_StartTagBegin = getBuffer().size();
        appendXML(Tag);
        m_StartTagEnd = getBuffer().size();
      }
      else
      {
        // There have been changes that require us to replace the XML text
        // that was already written.  This is a performance hit and
        // the programmer should minimize this activity as much as possible
        getBuffer().replace(m_StartTagBegin, m_StartTagEnd - m_StartTagBegin, Tag);
        m_StartTagEnd = m_StartTagBegin + Tag.size() - 1;
      }
      m_Dirty = false;
    }
  }

  virtual void popped(void)
  {
    _XMLStackString Tag;

    if ((m_StartTagBegin + m_StartTagEnd) == 0)
    {
      writeStartTag(true);
    }
    else
    {
      if (!m_Dirty)
      {
        // Check our attributes and make sure they aren't dirty
        for (AttributesVec::iterator Iter = m_AttributesVec.begin();
             Iter != m_AttributesVec.end(); ++Iter)
        {
          if ((*Iter)->isDirty())
          {
            m_Dirty = true;
            break;
          }
        }
      }

      if (m_Dirty)
      {
        // There have been changes that require us to replace the text
        // that was already written.  This is a performance hit and
        // the programmer should minimize this activity as much as possible
        writeStartTag(false);
      }
      Tag += m_Doc->VALS.LT + m_Doc->VALS.SLASH;
      Tag += m_Name;
      Tag += m_Doc->VALS.GT;
      appendXML(Tag);
    }
  }

  virtual _XMLStackNode * clone(void) const
  { 
    return new _XMLStackElement(*this); 
  }

  void checkName(void) const
  {
    // This is part of the set of characters that are illegal as a node
    // name.  In C++, to specify a non-printable character, you escape it
    // and provide it's octal code.  For example, "\13" is the decimal
    // character with the value 11.
    const wchar_t * IllegalChars = L"/;<=>?@[\\]^`\173\174\175\176";
    bool            Illegal      = false;
  
    if (m_Name.size() == 0)
    {
      throw XMLStackException("ERROR: Element nodes must have a defined"
        " name.");
    }

    // This performs some rudimentary checking for illegal node names.
    // This is by no means exhaustive, but does catch some of the more
    // common mistakes.
    // This is a time consuming test, so only perform it in debug builds
    int j;
    for (int i = 0; i < m_Name.size(); ++i)
    {
      if (m_Name[i] <= 44)
      {
        // No character below value 45 is legal
        Illegal = true;
      }

      j = 0;
      while (!Illegal && IllegalChars[j] != 0)
      {
        if (m_Name[i] == IllegalChars[j])
        {
          Illegal = true;
        }
        j++;
      }

      if (Illegal)
      {
        std::stringstream Msg;
        Msg << "ERROR: Invalid character found in node name";
        if (m_Name[i] < L' ' || m_Name[i] > L'z')
        {
          Msg << ".  Decimal value of illegal character is " << (int) m_Name[i];
        }
        else
        {
          Msg << " -->" << (char) m_Name[i] << "<--";
        }
        throw XMLStackException(Msg.str());
      }
    }
  }

};


/// Node class for creating attributes in element nodes
template<class _E, class _Tr = std::char_traits<_E>,  class _A = std::allocator<_E> >
class XMLStackAttribute : public XMLStackNode<_E, _Tr, _A>
{
public:
  /** Constructor which takes both an attribute name and an optional
   *  namespaceURI
   *
   *  The attribute name must conform to the rules for valid element
   *  names.  Only some basic checking is done to make sure the
   *  name is valid, and only when the _DEBUG preprocessor symbol is
   *  defined.
   *
   *  If a NamespaceURI is provided, no checking is performed to see
   *  if it was declared in an ancestor element; it will be written
   *  out regardless.  If you want the attribute in a namespace but don't
   *  want the xmlns declaration again, only prefix the attribute name
   *  with the namespace qualifier (e.g., "NamespaceQualifier:AttributeName").
   *  Don't provide the NamespaceURI.
   *
   *  \param Doc The XMLStackDocument this attribute belongs to
   *  \param Name The name of the attribute, along with any namespace prefix.
   *  \param NamespaceURI Optional string defining the namespace URI.  The
   *         attribute name must have a namespace prefix.  Attributes
   *         cannot be scoped to a default namespace.
   *
   *  \exception XMLStackException Thrown if Name is invalid
   */
  XMLStackAttribute(_XMLStackDocument *     Doc,
                    const _XMLStackString & Name,
                    const _XMLStackString & NamespaceURI = _XMLStackString())
   : _XMLStackNode(Doc, Name), m_NamespaceURI(NamespaceURI)
  {
    m_Dirty = false;

    // This is a time consuming test, so only perform it in debug builds
#ifdef _DEBUG
    // This is part of the set of characters that are illegal as a node
    // name.  In C++, to specify a non-printable character, you escape it
    // and provide it's octal code.  For example, "\13" is the decimal
    // character with the value 11.
    const wchar_t * IllegalChars = L"/;<=>?@[\\]^`\173\174\175\176";
    bool            Illegal      = false;
  
    if (Name.size() == 0)
    {
      throw XMLStackException("ERROR: Attribute nodes must have a defined"
        " name.");
    }

    // This performs some rudimentary checking for illegal node names.
    // This is by no means exhaustive, but does catch some of the more
    // common mistakes.
    int j;
    for (int i = 0; i < Name.size(); ++i)
    {
      if (Name[i] <= 44)
      {
        // No character below value 45 is legal
        Illegal = true;
      }

      j = 0;
      while (!Illegal && IllegalChars[j] != L'\0')
      {
        if (Name[i] == IllegalChars[j])
        {
          Illegal = true;
        }
        j++;
      }

      if (Illegal)
      {
        std::stringstream Msg;
        Msg << "ERROR: Invalid character found in node name";
        if (Name[i] < L' ' || Name[i] > L'z')
        {
          Msg << ".  Decimal value of illegal character is " << (int) Name[i];
        }
        else
        {
          Msg << " -->" << (char) Name[i] << "<--";
        }
        throw XMLStackException(Msg.str());
      }
    }
#endif
  }
  
  /// Destructor
  virtual ~XMLStackAttribute() {}

  /// Returns the URI for the namespace (if any)
  const _XMLStackString & getNamespaceURI(void) const
  {
    return m_NamespaceURI;
  }

  /** Sets the value for the attribute node
   *
   *  Automatic conversion is performed for characters that must be escaped.
   */
  virtual void setValue(const _XMLStackString & Value)
  {
    m_Value = Value;
    m_Dirty = true;
  }
  virtual void setValue(unsigned long Value) { setValue(convertValue(Value)); }
  virtual void setValue(long Value) { setValue(convertValue(Value)); }
  virtual void setValue(unsigned int Value) { setValue(convertValue(Value)); }
  virtual void setValue(int Value) { setValue(convertValue(Value)); }
  virtual void setValue(unsigned short Value) { setValue(convertValue(Value)); }
  virtual void setValue(short Value) { setValue(convertValue(Value)); }
  virtual void setValue(signed char Value) { setValue(convertValue(Value)); }
  virtual void setValue(char Value) { setValue(convertValue(Value)); }
  virtual void setValue(unsigned char Value) { setValue(convertValue(Value)); }
  virtual void setValue(double Value, int Scale = 6)
  { 
    setValue(convertValue(Value, Scale)); 
  }
  virtual void setValue(float Value, int Scale = 6)
  { 
    setValue(convertValue(Value, Scale)); 
  }

  
  /// Returns the XML for the attribute node
  _XMLStackString xml(void) const
  {
    _XMLStackString Value;

    if (m_NamespaceURI.length() > 0)
    {
      _XMLStackString Prefix = getPrefix();

      if (Prefix.length() > 0)
      {
        Value += m_Doc->VALS.XMLNS + m_Doc->VALS.COLON;
        Value += Prefix + m_Doc->VALS.EQUAL + m_Doc->VALS.QUOTE;
        Value += m_NamespaceURI + m_Doc->VALS.QUOTE + m_Doc->VALS.SPACE;
      }
    }

    Value += m_Name + m_Doc->VALS.EQUAL + m_Doc->VALS.QUOTE;

    if (m_Value.size() > 0)
    {
      _XMLStackString::size_type Pos = 0;

      Pos = Value.size();
      Value += m_Value;

      while (Pos < Value.size())
      {
        if (Value[Pos] == '&')
        {
          Value.replace(Pos, 1, m_Doc->VALS.AMP_REF);
          Pos += m_Doc->VALS.AMP_REF.size();
        }
        else if (Value[Pos] == '<')
        {
          Value.replace(Pos, 1, m_Doc->VALS.LT_REF);
          Pos += m_Doc->VALS.LT_REF.size();
        }
        else if (Value[Pos] == '"')
        {
          Value.replace(Pos, 1, m_Doc->VALS.QUOTE_REF);
          Pos += m_Doc->VALS.QUOTE_REF.size();
        }
        else
        {
          Pos++;
        }
      }
    }
    Value += m_Doc->VALS.QUOTE;

    return Value;
  }

  /// Returns true if the node has been modified
  bool isDirty(void) const { return m_Dirty; }

  /// Sets the dirty flag to the value passed in
  void setDirty(bool Dirty) { m_Dirty = Dirty; }

  /// Convenience operator for setValue()
  _XMLStackAttribute & operator=(const _XMLStackString & Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackAttribute & operator=(unsigned long & Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackAttribute & operator=(long Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackAttribute & operator=(unsigned int Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackAttribute & operator=(int Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackAttribute & operator=(unsigned short Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackAttribute & operator=(short Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackAttribute & operator=(signed char Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackAttribute & operator=(char Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackAttribute & operator=(unsigned char Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackAttribute & operator=(double Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackAttribute & operator=(float Value)
  {
    setValue(Value);
    return *this;
  }


protected:

  _XMLStackString m_NamespaceURI;
  _XMLStackString m_Value;
  bool            m_Dirty;

  virtual _XMLStackNode * clone(void) const
  { 
    return new _XMLStackAttribute(*this); 
  }
  friend class XMLStackElement<_E, _Tr, _A>;
};


/** This node creates textual data inside the XML.
 *
 *  Since several characters are used for markup (such as the ampersand &),
 *  this class will automatically scan the text and replace the characters
 *  with the proper entity reference.
 */
template<class _E, class _Tr = std::char_traits<_E>,  class _A = std::allocator<_E> >
class XMLStackText : public XMLStackNode<_E, _Tr, _A>
{
public:
  typedef XMLStackText<_E, _Tr, _A> _XMLStackText;

  /** Constructor
   *  
   *  \param Doc The XMLStackDocument instance this node belongs to.
   *  \param Data The textual data that will be written out to the XML
   *         buffer.  Any markup characters will be automatically
   *         replaced with the appropriate entity reference.
   */
  XMLStackText(_XMLStackDocument *     Doc,
               const _XMLStackString & Data)
   : _XMLStackNode(Doc, _XMLStackString()), m_Value(Data)
  {
  }

  /// Destructor
  virtual ~XMLStackText() {}

  /// Replaces the textual data defined in the constructor.
  virtual void setValue(const _XMLStackString & Value) { m_Value = Value; }
  virtual void setValue(unsigned long Value) { setValue(convertValue(Value)); }
  virtual void setValue(long Value) { setValue(convertValue(Value)); }
  virtual void setValue(unsigned int Value) { setValue(convertValue(Value)); }
  virtual void setValue(int Value) { setValue(convertValue(Value)); }
  virtual void setValue(unsigned short Value) { setValue(convertValue(Value)); }
  virtual void setValue(short Value) { setValue(convertValue(Value)); }
  virtual void setValue(signed char Value) { setValue(convertValue(Value)); }
  virtual void setValue(char Value) { setValue(convertValue(Value)); }
  virtual void setValue(unsigned char Value) { setValue(convertValue(Value)); }
  virtual void setValue(double Value, int Scale = 6)
  { 
    setValue(convertValue(Value, Scale)); 
  }
  virtual void setValue(float Value, int Scale = 6)
  { 
    setValue(convertValue(Value, Scale)); 
  }
  
  /// Convenience operator for setValue()
  _XMLStackText & operator=(const _XMLStackString & Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackText & operator=(unsigned long & Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackText & operator=(long Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackText & operator=(unsigned int Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackText & operator=(int Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackText & operator=(unsigned short Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackText & operator=(short Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackText & operator=(signed char Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackText & operator=(char Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackText & operator=(unsigned char Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackText & operator=(double Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackText & operator=(float Value)
  {
    setValue(Value);
    return *this;
  }

protected:
  _XMLStackString m_Value;

  virtual void popped(void)
  {
    if (m_Value.size() == 0)
    {
      return;
    }
    _XMLStackString::size_type Pos = 0;

    while (Pos < m_Value.size())
    {
      if (m_Value[Pos] == '&')
      {
        m_Value.replace(Pos, 1, m_Doc->VALS.AMP_REF);
        Pos += m_Doc->VALS.AMP_REF.size();
      }
      else if (m_Value[Pos] == '<')
      {
        m_Value.replace(Pos, 1, m_Doc->VALS.LT_REF);
        Pos += m_Doc->VALS.LT_REF.size();
      }
      else
      {
        Pos++;
      }
    }
    appendXML(m_Value);
  }
  
  virtual _XMLStackNode * clone(void) const
  { 
    return new _XMLStackText(*this); 
  }
};

/** CDATA sections are not parsed and thus can contain markup
 *
 *  Like text nodes, CDATA sections contain raw data.  However,
 *  the special characters '<' and '&' do not need to be escaped.
 *  CDATA sections can be used to quote or escape blocks of text
 *  to keep that text from being interpreted as markup language.
 *  The only delimiter that CDATA recognizes is "]]>", which
 *  ends the CDATA section.  Unfortunately, if this delimiter
 *  can possibly appear in your data, you will need to use
 *  a XMLStackText node, at least for part of the data.
 */
template<class _E, class _Tr = std::char_traits<_E>,  class _A = std::allocator<_E> >
class XMLStackCDATASection : public XMLStackNode<_E, _Tr, _A>
{
public:
  typedef XMLStackCDATASection<_E, _Tr, _A> _XMLStackCDATASection;

  /** Constructor
   *  
   *  \param Doc The XMLStackDocument instance this node belongs to.
   *  \param Data The textual data that will be written out to the XML
   *         buffer.
   *
   *  \exception XMLStackException Thrown if the delimiter "]]>" appears
   *             in the text and _DEBUG is defined.
   */
  XMLStackCDATASection(_XMLStackDocument *     Doc,
                       const _XMLStackString & Data)
   : _XMLStackNode(Doc, _XMLStackString()), m_Value(Data)
  {
  }

  /// Destructor
  virtual ~XMLStackCDATASection() {}

  /// Replaces the textual data defined in the constructor.
  virtual void setValue(const _XMLStackString & Value) { m_Value = Value; }
  virtual void setValue(unsigned long Value) { setValue(convertValue(Value)); }
  virtual void setValue(long Value) { setValue(convertValue(Value)); }
  virtual void setValue(unsigned int Value) { setValue(convertValue(Value)); }
  virtual void setValue(int Value) { setValue(convertValue(Value)); }
  virtual void setValue(unsigned short Value) { setValue(convertValue(Value)); }
  virtual void setValue(short Value) { setValue(convertValue(Value)); }
  virtual void setValue(signed char Value) { setValue(convertValue(Value)); }
  virtual void setValue(char Value) { setValue(convertValue(Value)); }
  virtual void setValue(unsigned char Value) { setValue(convertValue(Value)); }
  virtual void setValue(double Value, int Scale = 6)
  { 
    setValue(convertValue(Value, Scale)); 
  }
  virtual void setValue(float Value, int Scale = 6)
  { 
    setValue(convertValue(Value, Scale)); 
  }
  
  /// Convenience operator for setValue()
  _XMLStackCDATASection & operator=(const _XMLStackString & Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackCDATASection & operator=(unsigned long & Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackCDATASection & operator=(long Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackCDATASection & operator=(unsigned int Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackCDATASection & operator=(int Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackCDATASection & operator=(unsigned short Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackCDATASection & operator=(short Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackCDATASection & operator=(signed char Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackCDATASection & operator=(char Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackCDATASection & operator=(unsigned char Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackCDATASection & operator=(double Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackCDATASection & operator=(float Value)
  {
    setValue(Value);
    return *this;
  }

protected:
  _XMLStackString m_Value;

  virtual void popped(void)
  {
#ifdef _DEBUG
    if (m_Value.find(m_Doc->CDATA_END) != _XMLStackString::npos)
    {
      // Oops, "]]>" is the only reserved delimiter for CDATASection sections
      // and it cannot be escaped.  Throw an error.
      throw XMLStackException("ERROR: CDATASection node contains the reserved"
        " delimiter \"]]>\", which cannot be escaped.");
    }
#endif
    _XMLStackString Value;

    Value = m_Doc->CDATA_START;
    Value += m_Value;
    Value += m_Doc->CDATA_END;
    appendXML(Value);
  }

  virtual _XMLStackNode * clone(void) const
  { 
    return new _XMLStackCDATASection(*this); 
  }
};


/** Represents a comment in the XML
 *
 *  Comments are not interpreted by XML parsers.  The only requirement is
 *  that the character sequence "--" does not appear in the comment,
 *  since that is used for the ending delimiter.  Along the same lines,
 *  the comment cannot end with a single dash - either, since that will
 *  interfere with the ending delimiter.
 */
template<class _E, class _Tr = std::char_traits<_E>,  class _A = std::allocator<_E> >
class XMLStackComment : public XMLStackNode<_E, _Tr, _A>
{
public:
  typedef XMLStackComment<_E, _Tr, _A> _XMLStackComment;

  /** Constructor
   *  
   *  \param Doc The XMLStackDocument instance this node belongs to.
   *  \param Data The textual data that will be written out to the XML
   *         buffer as a comment.
   *
   *  \exception XMLStackException Thrown if the delimiter "--" appears
   *             in the text or the comment sends with a single dash
   *             and _DEBUG is defined.
   */
  XMLStackComment(_XMLStackDocument *     Doc,
                  const _XMLStackString & Data)
   : _XMLStackNode(Doc, _XMLStackString()), m_Text(Data)
  {
  }

  /// Destructor
  virtual ~XMLStackComment() {}

  /// Replaces the textual data defined in the constructor.
  virtual void setValue(const _XMLStackString & Value) { m_Text = Value; }
  virtual void setValue(unsigned long Value) { setValue(convertValue(Value)); }
  virtual void setValue(long Value) { setValue(convertValue(Value)); }
  virtual void setValue(unsigned int Value) { setValue(convertValue(Value)); }
  virtual void setValue(int Value) { setValue(convertValue(Value)); }
  virtual void setValue(unsigned short Value) { setValue(convertValue(Value)); }
  virtual void setValue(short Value) { setValue(convertValue(Value)); }
  virtual void setValue(signed char Value) { setValue(convertValue(Value)); }
  virtual void setValue(char Value) { setValue(convertValue(Value)); }
  virtual void setValue(unsigned char Value) { setValue(convertValue(Value)); }
  virtual void setValue(double Value, int Scale = 6)
  { 
    setValue(convertValue(Value, Scale)); 
  }
  virtual void setValue(float Value, int Scale = 6)
  { 
    setValue(convertValue(Value, Scale)); 
  }
  
  /// Convenience operator for setValue()
  _XMLStackComment & operator=(const _XMLStackString & Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackComment & operator=(unsigned long & Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackComment & operator=(long Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackComment & operator=(unsigned int Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackComment & operator=(int Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackComment & operator=(unsigned short Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackComment & operator=(short Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackComment & operator=(signed char Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackComment & operator=(char Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackComment & operator=(unsigned char Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackComment & operator=(double Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackComment & operator=(float Value)
  {
    setValue(Value);
    return *this;
  }

protected:
  _XMLStackString m_Text;

  virtual void popped(void)
  {
#ifdef _DEBUG
    if (m_Text.find(m_Doc->VALS.DASH2) != _XMLStackString::npos
      || (m_Text.size() > 0 && m_Text[m_Text.size() - 1] == '-'))
    {
      throw XMLStackException("ERROR: Comment nodes cannot contain the"
        " character sequence \"--\" or end with a single dash '-'");
    }
#endif
    _XMLStackString Value(m_Doc->VALS.COMMENT_START);
    Value += m_Text + m_Doc->VALS.COMMENT_END;
    appendXML(Value);
  }
  virtual _XMLStackNode * clone(void) const
  { 
    return new _XMLStackComment(*this); 
  }
};


/// Represents a processor instruction
template<class _E, class _Tr = std::char_traits<_E>,  class _A = std::allocator<_E> >
class XMLStackProcessingInstruction : public XMLStackNode<_E, _Tr, _A>
{
public:
  typedef XMLStackProcessingInstruction<_E, _Tr, _A>
    _XMLStackProcessingInstruction;
  
  /** Constructor
   *  
   *  \param Doc The XMLStackDocument instance this node belongs to.
   *  \param Target The name of the application target this PI is directed
   *
   *  \exception XMLStackException Thrown if Target is empty and _DEBUG
   *             is defined.
   */
  XMLStackProcessingInstruction(_XMLStackDocument *     Doc,
                                const _XMLStackString & Target)
   : _XMLStackNode(Doc, Target)
  {
#ifdef _DEBUG
    if (Target.size() == 0)
    {
      throw XMLStackException("ERROR: Processing Instruction nodes must"
        " have a defined name.");
    }
#endif 
  }

  /// Destructor
  virtual ~XMLStackProcessingInstruction() {}

  /// Replaces the content of the processing instruction
  virtual void setValue(const _XMLStackString & Value) { m_Value = Value; }
  virtual void setValue(unsigned long Value) { setValue(convertValue(Value)); }
  virtual void setValue(long Value) { setValue(convertValue(Value)); }
  virtual void setValue(unsigned int Value) { setValue(convertValue(Value)); }
  virtual void setValue(int Value) { setValue(convertValue(Value)); }
  virtual void setValue(unsigned short Value) { setValue(convertValue(Value)); }
  virtual void setValue(short Value) { setValue(convertValue(Value)); }
  virtual void setValue(signed char Value) { setValue(convertValue(Value)); }
  virtual void setValue(char Value) { setValue(convertValue(Value)); }
  virtual void setValue(unsigned char Value) { setValue(convertValue(Value)); }
  virtual void setValue(double Value, int Scale = 6)
  { 
    setValue(convertValue(Value, Scale)); 
  }
  virtual void setValue(float Value, int Scale = 6)
  { 
    setValue(convertValue(Value, Scale)); 
  }
  
  /// Convenience operator for setValue()
  _XMLStackProcessingInstruction & operator=(const _XMLStackString & Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackProcessingInstruction & operator=(unsigned long & Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackProcessingInstruction & operator=(long Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackProcessingInstruction & operator=(unsigned int Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackProcessingInstruction & operator=(int Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackProcessingInstruction & operator=(unsigned short Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackProcessingInstruction & operator=(short Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackProcessingInstruction & operator=(signed char Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackProcessingInstruction & operator=(char Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackProcessingInstruction & operator=(unsigned char Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackProcessingInstruction & operator=(double Value)
  {
    setValue(Value);
    return *this;
  }

  /// Convenience operator for setValue()
  _XMLStackProcessingInstruction & operator=(float Value)
  {
    setValue(Value);
    return *this;
  }

protected:
  _XMLStackString m_Value;

  virtual void popped(void)
  {
    _XMLStackString Value(m_Doc->VALS.PI_START);
    Value += m_Name + m_Doc->VALS.SPACE + m_Value + m_Doc->VALS.PI_END;
    appendXML(Value);
  }

  virtual _XMLStackNode * clone(void) const
  { 
    return new _XMLStackProcessingInstruction(*this); 
  }
};


/** This node isn't really a node per se, but a way to write XML directly
 *  into the XML buffer.
 *
 *  One potential use for this node is combining one or more XMLStackDocuments
 *  into a larger XML document.  If different portions of a larger XML
 *  document are created on several threads, then this node type could
 *  be used to create the master XML document.
 *
 *  Another use could be for document type declaration support.  Since DTDs
 *  aren't explicitly supported, this node type will allow you to add
 *  those declarations into the document.
 */
template<class _E, class _Tr = std::char_traits<_E>,  class _A = std::allocator<_E> >
class XMLStackRawXML : public XMLStackNode<_E, _Tr, _A>
{
public:
  typedef XMLStackRawXML<_E, _Tr, _A> _XMLStackRawXML;

  /** Constructor
   *
   *  \param PushXML The XML that is written to the document's XML buffer
   *         immediately.  Can be an empty string.
   *  \param PopXML The XML that will be written to the document's XML
   *         buffer when the node is popped off.  Can be an empty string.
   */
  XMLStackRawXML(_XMLStackDocument *     Doc,
                 const _XMLStackString & PushXML,
                 const _XMLStackString & PopXML)
   : _XMLStackNode(Doc, _XMLStackString()), m_PopXML(PopXML)
  {
    appendXML(PushXML);
  }

  /// Destructor
  virtual ~XMLStackRawXML() {}

protected:
  _XMLStackString m_PopXML;

  virtual bool childNodePushed(void)
  {
    return true;
  }

  virtual void popped(void)
  {
    appendXML(m_PopXML);
  }

  virtual _XMLStackNode * clone(void) const
  { 
    return new _XMLStackRawXML(*this); 
  }
};


/// Defines string constants needed by the node classes
template<class _E, class _Tr = std::char_traits<_E>,  class _A = std::allocator<_E> >
struct XMLStringConstants
{
  typedef std::basic_string<_E, _Tr, _A> _XMLStackString;

  XMLStringConstants()
    : LT(_CE(L"<")),
      GT(_CE(L">")),
      SPACE(_CE(L" ")),
      XMLNS(_CE(L"xmlns")),
      S_XMLNS_C(_CE(L" xmlns:")),
      COLON(_CE(L":")),
      EQUAL(_CE(L"=")),
      QUOTE(_CE(L"\"")),
      AMP_REF(_CE(L"&amp;")),
      QUOTE_REF(_CE(L"&quot;")),
      LT_REF(_CE(L"&lt;")),
      SLASH(_CE(L"/")),
      CDATA_START(_CE(L"<![CDATA[")),
      CDATA_END(_CE(L"]]>")),
      COMMENT_START(_CE(L"<!--")),
      COMMENT_END(_CE(L"-->")),
      DASH2(_CE(L"--")),
      PI_START(_CE(L"<?")),
      PI_END(_CE(L"?>"))
  {
  }

  // These string declarations minimize the impact of constantly
  // converting to the correct string type
  const _XMLStackString LT;
  const _XMLStackString GT;
  const _XMLStackString SPACE;
  const _XMLStackString XMLNS;
  const _XMLStackString S_XMLNS_C;
  const _XMLStackString COLON;
  const _XMLStackString EQUAL;
  const _XMLStackString QUOTE;
  const _XMLStackString AMP_REF;
  const _XMLStackString QUOTE_REF;
  const _XMLStackString LT_REF;
  const _XMLStackString SLASH;
  const _XMLStackString CDATA_START;
  const _XMLStackString CDATA_END;
  const _XMLStackString COMMENT_START;
  const _XMLStackString COMMENT_END;
  const _XMLStackString DASH2;
  const _XMLStackString PI_START;
  const _XMLStackString PI_END;
};

/** XMLStackDocument is a fast stack-based XML writer
 *
 *  This class creates XML by pushing and popping various node types
 *  onto an internal stack.  As nodes are pushed and popped, an internal
 *  buffer holds the XML created by the nodes.  
 *
 *  The XML is created in a std::basic_string class.  The classes
 *  are written as templates so that they can support either ANSI or Unicode
 *  type strings.  Two typedefs are been defined for convenience:
 *  AXMLStackDocument (for ANSI strings) and WXMLStackDocument (for Unicode).
 *
 *  \author Scott Moore
 *  \version 1.0
 *  \date July 2001
 *
 *  Here's a sample XML file and the XMLStack code used to create it:
 *
\code
<?xml version="1.0" encoding="UTF-16"?>
<Book Status="BackOrdered">
  <Title>Net Etiquette</Title>
  <PubId>736</PubId>
  <Price Currency="USD">19.99</Price>
  <Author>
    <LastName>White</LastName>
    <FirstName>Johnson</FirstName>
  </Author>
</Book>


WXMLStackDocument Doc;  // Unicode version

Doc.pushProcessingInstruction(L"xml") = L"version='1.0' encoding='UTF-16'";
Doc.pop();  // pops the processing instruction

Doc.pushElement(L"Book");
Doc.top().createAttribute(L"Status") = L"BackOrdered";

Doc.pushElement(L"Title") = L"Net Etiquette";
Doc.pop();  // pops <Title>

Doc.pushElement(L"PubId") = 736;
Doc.pop();  // pops <PubId>

Doc.pushElement(L"Price").createAttribute(L"Currency") = L"USD";
Doc.top().setValue(19.99, 2);  // Assign 19.99 to <Price>
Doc.pop();  // pops <Price>

Doc.pushElement(L"Author");
Doc.pushElement(L"LastName") = L"White";
Doc.pop();  // pops <LastName>

Doc.pushElement(L"FirstName").setValue(L"Johnson");
Doc.popAll(); // Pops off all remaining nodes
\endcode
 *
 * Here's another example which demostrates the namespace capabilities:
 *
\code
<SOAP-ENV:Envelope xmlns:SOAP-ENV="http://schemas.xmlsoap.org/soap/envelope/"
SOAP-ENV:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/" 
xmlns:TS="http://www.time.gov/">
  <SOAP-ENV:Header/>
  <SOAP-ENV:Body>
    <TS:GetCurrentTimeRequest>
      <TimeZone>Eastern</TimeZone>
    </TS:GetCurrentTimeRequest>
  </SOAP-ENV:Body>
</SOAP-ENV:Envelope>


AXMLStackDocument Doc;  // ANSI version

Doc.pushElement("SOAP-ENV:Envelope", 
  "http://schemas.xmlsoap.org/soap/envelope/");
Doc.top().createAttribute("SOAP-ENV:encodingStyle") =
  "http://schemas.xmlsoap.org/soap/encoding/";
Doc.top().createAttribute("xmlns:TS") = "http://www.time.gov/";

Doc.pushElement("SOAP-ENV:Header");
Doc.pop();

Doc.pushElement("SOAP-ENV:Body");

// Don't pass the namespace URI because it was already declared in an
// ancestor element.  We only need to place the namespace prefix TS
// in front of the node name to place it in the "http://www.time.gov/"
// namespace.
Doc.pushElement("TS:GetCurrentTimeRequest");

Doc.pushElement("TimeZone") = "Eastern";
Doc.popAll(); // Pops off all remaining nodes
\endcode
 */ 
template<class _E, class _Tr = std::char_traits<_E>,  class _A = std::allocator<_E> >
class XMLStackDocument
{
public:
  typedef std::basic_string<_E, _Tr, _A> _XMLStackString;
  typedef XMLStackDocument<_E, _Tr, _A> _XMLStackDocument;
  typedef XMLStackNode<_E, _Tr, _A> _XMLStackNode;
  typedef XMLStackElement<_E, _Tr, _A> _XMLStackElement;
  typedef XMLStackText<_E, _Tr, _A> _XMLStackText;
  typedef XMLStackCDATASection<_E, _Tr, _A> _XMLStackCDATASection;
  typedef XMLStackComment<_E, _Tr, _A> _XMLStackComment;
  typedef XMLStackProcessingInstruction<_E, _Tr, _A> _XMLStackProcessingInstruction;
  typedef XMLStackRawXML<_E, _Tr, _A> _XMLStackRawXML;
  
  /** Constructor which takes an optional XML buffer size.  
   *
   *  \param InitialBufferSize The initial size of the XML buffer
   *         (in characters, not bytes).  See reserveXMLBufferSize()
   *         for details.
   */
  XMLStackDocument(unsigned long InitialBufferSize = 1024)
  {
    // Start us off with a decent size string to keep initial memory allocations
    // to a minimum.
    m_XML.reserve(InitialBufferSize);
  }

  /// Copy constructor
  XMLStackDocument(const _XMLStackDocument & Instance)
  {
    *this = Instance;
  }

  /// Destructor
  ~XMLStackDocument()
  {
    freeStack();
  }

  /** Pops the top node off the stack.
   *
   *  Any XML that wasn't previously written to the buffer by
   *  the node will be written at this time.
   *
   *  \exception XMLStackException Thrown if no nodes are on the stack.
   */
  void pop(void)
  {
    if (m_Stack.size() == 0)
    {
      throw XMLStackException("ERROR: Pop() - XMLStackDocument is empty.");
    }
    m_Stack.back()->popped();
    delete m_Stack.back();
    m_Stack.pop_back();
  }

  /** Repeatedly calls pop() until no more nodes exist on the stack.
   *
   *  If no nodes exist on the stack, then this method does nothing.
   */
  void popAll(void)
  {
    while (m_Stack.size() > 0)
    {
      pop();
    }
  }
  
  /** Pushes an element node on the stack
   *
   *  The element name must conform to the rules for valid element
   *  names.  Only some basic checking is done to make sure the
   *  name is valid, and only when the _DEBUG preprocessor symbol is
   *  defined.
   *
   *  \param Name The name of the element node
   *
   *  \return Returns the instance of the XMLStackElement that was
   *          created.
   *
   *  \exception XMLStackException Thrown if Name is invalid or if the parent
   *             node does not allow children.
   */
  _XMLStackElement & pushElement(const _XMLStackString & Name)
  {
    if (m_Stack.size() > 0)
    {
      if (!m_Stack.back()->childNodePushed())
      {
        throw XMLStackException("ERROR: Cannot create a child XMLStackElement"
          " node of the given parent node.");
      }
    }
    _XMLStackElement * Element = new _XMLStackElement(this, Name);

    m_Stack.push_back(Element);
    return *Element;
  }
  
  /** Pushes an element node on the stack and declares the passed in namespace
   *
   *  The element name must conform to the rules for valid element
   *  names.  Only some basic checking is done to make sure the
   *  name is valid, and only when the _DEBUG preprocessor symbol is
   *  defined.
   *
   *  If a NamespaceURI is provided, no checking is performed to see
   *  if it was declared in an ancestor element; it will be written
   *  out regardless.  If you want the element in a namespace but don't
   *  want the xmlns declaration again, only prefix the element name
   *  with the namespace qualifier (e.g., "NamespaceQualifier:ElementName").
   *  Don't provide the NamespaceURI.
   *
   *  \param Name The name of the XML tag, with any namespace prefix
   *         included if present.
   *  \param NamespaceURI String defining the namespace URI. If specified,
   *         the node is created in the context of the namespaceURI
   *         with the prefix specified on the node name.  If the Name
   *         parameter does not have a prefix, this is treated as the
   *         default namespace. 
   *
   *  \return Returns the instance of the XMLStackElement that was
   *          created.
   *
   *  \exception XMLStackException Thrown if Name is invalid or if the parent
   *             node does not allow children.
   */
  _XMLStackElement & pushElement(const _XMLStackString & Name,
                                 const _XMLStackString & NamespaceURI)
  {
    if (m_Stack.size() > 0)
    {
      if (!m_Stack.back()->childNodePushed())
      {
        throw XMLStackException("ERROR: Cannot create a child XMLStackElement"
          " node of the given parent node.");
      }
    }
    _XMLStackElement * Element = new _XMLStackElement(this, Name, NamespaceURI);

    m_Stack.push_back(Element);
    return *Element;
  }
  
  /** Pushes a text node on the stack
   *
   *  Text nodes contain raw data that is placed in the XML stream.
   *  Automatic conversion is performed for the reserved characters '<'
   *  and '&'.  
   *
   *  \param Data The text that will be placed in the XML.
   *
   *  \return Returns the instance of the XMLStackText that was
   *          created.
   *
   *  \exception XMLStackException Thrown if the parent node does not allow
   *             children.
   */
  _XMLStackText & pushText(const _XMLStackString & Data)
  {
    if (m_Stack.size() > 0)
    {
      if (!m_Stack.back()->childNodePushed())
      {
        throw XMLStackException("ERROR: Cannot create a child XMLStackText"
          " node of the given parent node.");
      }
    }
    _XMLStackText * Text = new _XMLStackText(this, Data);

    m_Stack.push_back(Text);
    return *Text;
  }
  
  /** Pushes a CDATA section node on the stack
   *
   *  Like text nodes, CDATA sections contain raw data.  However,
   *  the special characters '<' and '&' do not need to be escaped.
   *  CDATA sections can be used to quote or escape blocks of text
   *  to keep that text from being interpreted as markup language.
   *  The only delimiter that CDATA recognizes is "]]>", which
   *  ends the CDATA section.  Unfortunately, if this delimiter
   *  can possibly appear in your data, you will need to use
   *  a XMLStackText node, at least for part of the data.
   *
   *  \param Data The text that will be placed in the XML.
   *
   *  \return Returns the instance of the XMLStackCDATASection that was
   *          created.
   *
   *  \exception XMLStackException Thrown if the delimiter "]]>" exists
   *             in Data or if the parent node cannot have child nodes.
   */
  _XMLStackCDATASection & pushCDATASection(const _XMLStackString & Data)
  {
    if (m_Stack.size() > 0)
    {
      if (!m_Stack.back()->childNodePushed())
      {
        throw XMLStackException("ERROR: Cannot create a child XMLStackCDATASection"
          " node of the given parent node.");
      }
    }
    _XMLStackCDATASection * Text = new _XMLStackCDATASection(this, Data);

    m_Stack.push_back(Text);
    return *Text;
  }
  
  /** Pushes a comment node on the stack
   *
   *  Comments cannot contain the character sequence "--".
   *
   *  \param Data The text that will be placed in the XML between
   *         the comment delimiters <!-- and -->.
   *
   *  \return Returns the instance of the XMLStackComment that was
   *          created.
   *
   *  \exception XMLStackException Thrown if the delimiter "--" exists
   *             in Data or if the parent node cannot have child nodes.
   */
  _XMLStackComment & pushComment(const _XMLStackString & Data)
  {
    if (m_Stack.size() > 0)
    {
      if (!m_Stack.back()->childNodePushed())
      {
        throw XMLStackException("ERROR: Cannot create a child XMLStackComment"
          " node of the given parent node.");
      }
    }
    _XMLStackComment * Comment = new _XMLStackComment(this, Data);

    m_Stack.push_back(Comment);
    return *Comment;
  }
  
  /** Pushes a processing instruction node on the stack
   *
   *  Represents a processing instruction, which XML defines to keep
   *  processor-specific information in the text of the document.
   *
   *  \param Target The application to which this processing instruction
   *         is directed
   *
   *  \return Returns the instance of the XMLStackProcessingInstruction
   *          that was created.
   *
   *  \exception XMLStackException Thrown if Target is empty or the parent
   *             node cannot have child nodes.
   */
  _XMLStackProcessingInstruction & pushProcessingInstruction
    (const _XMLStackString & Target)
  {
    if (m_Stack.size() > 0)
    {
      if (!m_Stack.back()->childNodePushed())
      {
        throw XMLStackException("ERROR: Cannot create a child XMLStackProcessingInstruction"
          " node of the given parent node.");
      }
    }
    _XMLStackProcessingInstruction * PI =
      new _XMLStackProcessingInstruction(this, Target);

    m_Stack.push_back(PI);
    return *PI;
  }
  
  /** Pushes a raw XML node on the stack
   *
   *  This node is used when you want to add raw XML to the buffer
   *  without any processing performed on it.  Other nodes can be pushed
   *  on top of this node type (e.g., child nodes).
   *
   *  \param PushXML This is the XML that is added to the buffer
   *         immediately (during the push operation).  It can be
   *         empty.
   *  \param PopXML This XML is added to the buffer at the time this
   *         node is popped off the stack.  It can be empty.
   *
   *  \return Returns the instance of the XMLStackRawXML that was created.
   *
   *  \exception XMLStackException Thrown if the parent node cannot have child
   *             nodes.
   */
  _XMLStackRawXML & pushRawXML(const _XMLStackString & PushXML,
                               const _XMLStackString & PopXML)
  {
    if (m_Stack.size() > 0)
    {
      if (!m_Stack.back()->childNodePushed())
      {
        throw XMLStackException("ERROR: Cannot create a child XMLStackRawXML"
          " node of the given parent node.");
      }
    }
    _XMLStackRawXML * Node = new _XMLStackRawXML(this, PushXML, PopXML);

    m_Stack.push_back(Node);
    return *Node;
  }

  /** Returns the node on top of the stack.
   *
   *  \exception XMLStackException Thrown if stackSize() == 0
   */
  _XMLStackNode & top(void)
  {
    if (stackSize() > 0)
    {
      return *m_Stack.back();
    }
    throw XMLStackException("ERROR: top() called on empty stack.");
  }

  /** Returns the node on top of the stack.
   *
   *  \exception XMLStackException Thrown if stackSize() == 0
   */
  const _XMLStackNode & top(void) const
  {
    if (stackSize() > 0)
    {
      return *m_Stack.back();
    }
    throw XMLStackException("ERROR: top() called on empty stack.");
  }

  /** Returns the contents of the XML buffer.
   *
   *  \note If nodes are still on the stack, the XML will probably
   *        not be well-formed.
   */
  const _XMLStackString & xml(void) const { return m_XML; }

  /** Clears the stack, erasing all XML and deleting all nodes.
   *
   *  \note This will reset the internal XML buffer to a default size.
   *        Call reserveXMLBufferSize() to allocate additional
   *        buffer memory if desired.
   */
  void clear(void)
  {
    freeStack();
    m_XML.erase();
  }

  /// Returns the number of nodes currently on the stack.  Zero if empty.
  unsigned long stackSize(void) const { return m_Stack.size(); }

  /** Creates a buffer for the XML at least as large as Capacity.
   *
   *  This is optional and purely for performance.  If you have a good idea
   *  of the final buffer size, setting this value early on will give
   *  better performance.  Otherwise, the buffer will grow as needed.
   *
   *  \param Capacity The size of the buffer (in characters, not bytes)
   *         to reserve.  If Capacity is smaller than the current size,
   *         then this method does nothing.
   */
  void reserveXMLBufferSize(unsigned long Capacity)
  {
    if (m_XML.capacity() < Capacity)
    {
      m_XML.reserve(Capacity);
    }
  }

  /// Makes a copy of the passed in XMLStackDocument instance.
  _XMLStackDocument & operator=(const _XMLStackDocument & Instance)
  {
    m_XML = Instance.m_XML;
    freeStack();

    for (int i = 0; i < Instance.m_Stack.size(); ++i)
    {
      m_Stack.push_back(Instance.m_Stack[i]->clone());
      m_Stack.back()->m_Doc = this;
    }
    return *this;
  }

private:
  typedef std::vector<_XMLStackNode *> NodeStack;

  _XMLStackString                    m_XML;
  NodeStack                          m_Stack;
  const XMLStringConstants<_E, _Tr, _A> VALS;

  void appendXML(const _XMLStackString & XML)
  {
    if (XML.size() + m_XML.size() > m_XML.capacity())
    {
      // Microsoft's STL basic_string has a fixed sized increments implementation
      // of memory management.  This gets around that problem and provides
      // an enormous performance boost.
      m_XML.reserve(m_XML.capacity() * 2 + XML.size());
    }
    m_XML += XML;
  }

  void freeStack(void)
  {
    for (int i = 0; i < m_Stack.size(); ++i)
    {
      delete m_Stack[i];
    }
    m_Stack.clear();
  }

  friend class XMLStackNode<_E, _Tr, _A>;
  friend class XMLStackElement<_E, _Tr, _A>;
  friend class XMLStackAttribute<_E, _Tr, _A>;
  friend class XMLStackProcessingInstruction<_E, _Tr, _A>;
  friend class XMLStackText<_E, _Tr, _A>;
  friend class XMLStackCDATASection<_E, _Tr, _A>;
  friend class XMLStackRawXML<_E, _Tr, _A>;
  friend class XMLStackComment<_E, _Tr, _A>;
};


/// Specialization of XMLStringConstants for char
template<>
struct XMLStringConstants<char, std::char_traits<char>, std::allocator<char> >
{
  typedef std::basic_string<char, std::char_traits<char>, std::allocator<char> > _XMLStackString;

  XMLStringConstants()
    : LT("<"),
      GT(">"),
      SPACE(" "),
      XMLNS("xmlns"),
      S_XMLNS_C(" xmlns:"),
      COLON(":"),
      EQUAL("="),
      QUOTE("\""),
      AMP_REF("&amp;"),
      QUOTE_REF("&quot;"),
      LT_REF("&lt;"),
      SLASH("/"),
      CDATA_START("<![CDATA["),
      CDATA_END("]]>"),
      COMMENT_START("<!--"),
      COMMENT_END("-->"),
      DASH2("--"),
      PI_START("<?"),
      PI_END("?>")
  {
  }

  // These string declarations minimize the impact of constantly
  // converting to the correct string type
  const _XMLStackString LT;
  const _XMLStackString GT;
  const _XMLStackString SPACE;
  const _XMLStackString XMLNS;
  const _XMLStackString S_XMLNS_C;
  const _XMLStackString COLON;
  const _XMLStackString EQUAL;
  const _XMLStackString QUOTE;
  const _XMLStackString AMP_REF;
  const _XMLStackString QUOTE_REF;
  const _XMLStackString LT_REF;
  const _XMLStackString SLASH;
  const _XMLStackString CDATA_START;
  const _XMLStackString CDATA_END;
  const _XMLStackString COMMENT_START;
  const _XMLStackString COMMENT_END;
  const _XMLStackString DASH2;
  const _XMLStackString PI_START;
  const _XMLStackString PI_END;
};


/// Specialization of XMLStringConstants for wchar_t
template<>
struct XMLStringConstants<wchar_t, std::char_traits<wchar_t>, std::allocator<wchar_t> >
{
  typedef std::basic_string<wchar_t, std::char_traits<wchar_t>, std::allocator<wchar_t> > _XMLStackString;

  XMLStringConstants()
    : LT(L"<"),
      GT(L">"),
      SPACE(L" "),
      XMLNS(L"xmlns"),
      S_XMLNS_C(L" xmlns:"),
      COLON(L":"),
      EQUAL(L"="),
      QUOTE(L"\""),
      AMP_REF(L"&amp;"),
      QUOTE_REF(L"&quot;"),
      LT_REF(L"&lt;"),
      SLASH(L"/"),
      CDATA_START(L"<![CDATA["),
      CDATA_END(L"]]>"),
      COMMENT_START(L"<!--"),
      COMMENT_END(L"-->"),
      DASH2(L"--"),
      PI_START(L"<?"),
      PI_END(L"?>")
  {
  }

  // These string declarations minimize the impact of constantly
  // converting to the correct string type
  const _XMLStackString LT;
  const _XMLStackString GT;
  const _XMLStackString SPACE;
  const _XMLStackString XMLNS;
  const _XMLStackString S_XMLNS_C;
  const _XMLStackString COLON;
  const _XMLStackString EQUAL;
  const _XMLStackString QUOTE;
  const _XMLStackString AMP_REF;
  const _XMLStackString QUOTE_REF;
  const _XMLStackString LT_REF;
  const _XMLStackString SLASH;
  const _XMLStackString CDATA_START;
  const _XMLStackString CDATA_END;
  const _XMLStackString COMMENT_START;
  const _XMLStackString COMMENT_END;
  const _XMLStackString DASH2;
  const _XMLStackString PI_START;
  const _XMLStackString PI_END;
};


typedef XMLStackDocument<wchar_t, std::char_traits<wchar_t>,
  std::allocator<wchar_t> > WXMLStackDocument;
typedef XMLStackDocument<char, std::char_traits<char>,
  std::allocator<char> > AXMLStackDocument;

#endif // _XMLSTACK_H
